@@ -1,10 +1,9 @@
-import re
 import cv2
 import math
 import cv2 as cv
 import numpy as np
 import PIL.Image as Image
-
+import io
 import matplotlib
 import matplotlib.pyplot as plt
 
@@ -17,7 +16,7 @@ from ..utils import time_reduce
 from ..utils import glob_recursively
 from ..utils import glob_images
 from ..utils import has_chinese_char
-from ..utils import vvd_floor
+from ..utils import vvd_floor, vvd_ceil
 from ..utils import change_file_name_for_path
 from ..utils import encode_path
 from ..utils import cal_distance
@@ -414,6 +413,51 @@ def img_rotate(img,
 
 image_rotate = img_rotate
 
+def image_center_rotate(image, angle, center=None, interpolation=cv2.INTER_LINEAR, border_mode=cv2.BORDER_CONSTANT, border_value=0):
+    type = image.dtype
+    image = image.astype('uint8')
+    h, w = image.shape[:2]
+
+    if center is None:
+        center = ((w - 1) * 0.5, (h - 1) * 0.5)
+    assert isinstance(center, tuple)
+
+    matrix = cv2.getRotationMatrix2D((0, 0), -angle, 1)
+
+    corner_points = np.array([
+                        [0, 0],
+                        [w, 0],
+                        [w, h],
+                        [0, h]
+                        ] + [[center[0], center[1]]])
+
+    new_corner_points = corner_points @ matrix.T[:2, :]
+
+    x_min = np.min(new_corner_points[:, 0])
+    y_min = np.min(new_corner_points[:, 1])
+    x_max = np.max(new_corner_points[:, 0])
+    y_max = np.max(new_corner_points[:, 1])
+
+    matrix[0, 2] -= x_min
+    matrix[1, 2] -= y_min
+
+    new_w = np.ceil(x_max - x_min).astype('int32')
+    new_h = np.ceil(y_max - y_min).astype('int32')
+
+    new_corner_points -= [x_min, y_min]
+
+    rotated = cv2.warpAffine(
+        image,
+        matrix, (new_w, new_h),
+        flags=interpolation,
+        borderValue=border_value,
+        borderMode=border_mode)
+
+    rotated.astype(type)
+    center_point = new_corner_points[-1, :]
+
+    return rotated, center_point
+
 def image_flip_lr(image, contiguousarray=True):
     fliped = np.fliplr(image)
     if contiguousarray:
@@ -634,7 +678,7 @@ def image_show_from_path(file_path):
         plt_image_show(image)
 
 
-def plt_image_show(*image, window_name='', array_res=False, full_screen=True, cmap=None, position=[30, 30], share_xy=False, axis_off=False, col_num=None, row_num=None, norm_float=True):
+def plt_image_show(*image, window_name='', array_res=False, full_screen=True, cmap=None, position=[30, 30], share_xy=False, axis_off=False, col_num=None, row_num=None, norm_float=True, bgr=False):
     '''
     更加鲁棒地显示图像包括二维图像,第三维度为1的图像
     '''
@@ -690,6 +734,8 @@ def plt_image_show(*image, window_name='', array_res=False, full_screen=True, cm
             if image.shape[0] == 3 and image.shape[-1] != 3:
                 # guess it is a tensor image
                 image = np.transpose(image, [1, 2 ,0])
+            if image.shape[-1] == 3 and bgr:
+                image = cv_rgb_bgr_convert(image)
 
         if iterable(ax):
             if ax.ndim == 1:
@@ -793,7 +839,7 @@ def scatter3d(xyz_data_list, label_list, xyz_label=None, title='3d Scatter plot'
 
 
 def scatter2d(xy_data_list, label_list=None, xy_label=None, title='2d Scatter plot'):
-    if not xy_data_list:
+    if len(xy_data_list) == 0:
         return
     xyz_data = np.array(xy_data_list)
     assert xyz_data.ndim == 2
@@ -1232,6 +1278,13 @@ def polygon2bbox(polygon):
         array = array[:, 0, :]
     return np.min(array, axis=0).tolist() + np.max(array, axis=0).tolist()
 
+def ceil_bbox(bbox):
+    x1, y1, x2, y2 = bbox
+    return [vvd_floor(x1), vvd_floor(y1), vvd_ceil(x2), vvd_ceil(y2)]
+
+def floor_bbox(bbox):
+    x1, y1, x2, y2 = bbox
+    return [vvd_ceil(x1), vvd_ceil(y1), vvd_floor(x2), vvd_floor(y2)]
 
 def xyxy2polygon(x1, y1=None, x2=None, y2=None):
     if y1 is None and x2 is None and y2 is None:
@@ -1264,7 +1317,7 @@ def polygon_interpolation(polygon, max_distance, close=True):
     return points_res
 
 
-def get_polygons_from_mask(mask, approx=True, epsilon=None):
+def get_polygons_from_mask(mask, approx=False, epsilon=None):
     """ get polygons for all isolated areas on fg-mask
         return something like: [[x1,y1], [x2,y2], [x3,y3], ...]
     """
@@ -1793,6 +1846,72 @@ def mask_on_mask(input_mask, standard_mask):
     final_mask = (labels < 0).astype('uint8')
     return final_mask
 
+def img2video(image_list, output_video_name, fps, resize=None):
+    # 图像转视频 基于 OpenCV 
+    # 参数：image_list: rgb 图像列表，output_video_name: 输出视频文件名，fps: 视频帧率，resize: 图像尺寸调整，None 表示不调整
+
+    if resize is not None:
+        shape = (resize[0], resize[1])
+    else:
+        shape = None
+
+    # 获取图像的尺寸
+    image = image_list[0]
+    image = to_colorful_image(image)
+    if shape is not None:
+        image = image_resize(image, shape)
+
+    height, width, channels = image.shape
+    shape = (width, height)
+
+    # 创建视频 writer
+    fourcc = cv2.VideoWriter.fourcc(*'mp4v')  # 使用mp4视频编码
+    out = cv2.VideoWriter(output_video_name, fourcc, fps, (width, height))
+
+    # 遍历图像路径列表，并将图像写入视频文件
+    for image in tqdm(image_list, desc=f' @@ image to video running: '):
+        image = to_colorful_image(image)
+        image = cv_rgb_bgr_convert(image)
+        image = image_resize(image, shape)
+        out.write(image)
+
+    # 释放writer
+    out.release()
+
+
+def imgpath2video(image_path_list, output_video_name, fps, resize=None):
+    # 图像转视频 基于 OpenCV 
+    # 参数：image_path_list: 图像路径列表，output_video_name: 输出视频文件名，fps: 视频帧率，resize: 图像尺寸调整，None 表示不调整
+
+    if resize is not None:
+        shape = (resize[0], resize[1])
+    else:
+        shape = None
+
+    # 获取图像的尺寸
+    image = cv2.imread(image_path_list[0])
+    image = to_colorful_image(image)
+    if shape is not None:
+        image = image_resize(image, shape)
+
+    height, width, channels = image.shape
+    shape = (width, height)
+
+    # 创建视频 writer
+    fourcc = cv2.VideoWriter.fourcc(*'mp4v')  # 使用mp4视频编码
+    out = cv2.VideoWriter(output_video_name, fourcc, fps, (width, height))
+
+    # 遍历图像路径列表，并将图像写入视频文件
+    for image_path in tqdm(image_path_list, desc=f' @@ image to video running: '):
+        image = cv2.imread(image_path)
+        image = to_colorful_image(image)
+        image = image_resize(image, shape)
+        out.write(image)
+
+    # 释放writer
+    out.release()
+
+
 def distortion_calibration(image_dir_path, W_num, H_num, show=False):
     # termination criteria
     criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 30, 0.001)
@@ -1874,71 +1993,6 @@ def distortion_calibration(image_dir_path, W_num, H_num, show=False):
     return result
 
 
-def img2video(image_list, output_video_name, fps, resize=None):
-    # 图像转视频 基于 OpenCV 
-    # 参数：image_list: rgb 图像列表，output_video_name: 输出视频文件名，fps: 视频帧率，resize: 图像尺寸调整，None 表示不调整
-
-    if resize is not None:
-        shape = (resize[0], resize[1])
-    else:
-        shape = None
-
-    # 获取图像的尺寸
-    image = image_list[0]
-    image = to_colorful_image(image)
-    if shape is not None:
-        image = image_resize(image, shape)
-
-    height, width, channels = image.shape
-    shape = (width, height)
-
-    # 创建视频 writer
-    fourcc = cv2.VideoWriter.fourcc(*'mp4v')  # 使用mp4视频编码
-    out = cv2.VideoWriter(output_video_name, fourcc, fps, (width, height))
-
-    # 遍历图像路径列表，并将图像写入视频文件
-    for image in tqdm(image_list, desc=f' @@ image to video running: '):
-        image = to_colorful_image(image)
-        image = cv_rgb_bgr_convert(image)
-        image = image_resize(image, shape)
-        out.write(image)
-
-    # 释放writer
-    out.release()
-
-
-def imgpath2video(image_path_list, output_video_name, fps, resize=None):
-    # 图像转视频 基于 OpenCV 
-    # 参数：image_path_list: 图像路径列表，output_video_name: 输出视频文件名，fps: 视频帧率，resize: 图像尺寸调整，None 表示不调整
-
-    if resize is not None:
-        shape = (resize[0], resize[1])
-    else:
-        shape = None
-
-    # 获取图像的尺寸
-    image = cv2.imread(image_path_list[0])
-    image = to_colorful_image(image)
-    if shape is not None:
-        image = image_resize(image, shape)
-
-    height, width, channels = image.shape
-    shape = (width, height)
-
-    # 创建视频 writer
-    fourcc = cv2.VideoWriter.fourcc(*'mp4v')  # 使用mp4视频编码
-    out = cv2.VideoWriter(output_video_name, fourcc, fps, (width, height))
-
-    # 遍历图像路径列表，并将图像写入视频文件
-    for image_path in tqdm(image_path_list, desc=f' @@ image to video running: '):
-        image = cv2.imread(image_path)
-        image = to_colorful_image(image)
-        image = image_resize(image, shape)
-        out.write(image)
-
-    # 释放writer
-    out.release()
-
 def distortion_calibration_to_map(mtx, dist, W, H, **kwargs):
     result = dict()
 
@@ -1959,7 +2013,8 @@ def distortion_calibration_to_map(mtx, dist, W, H, **kwargs):
     # frame = cv2.remap(img, map1, map2, cv2.INTER_LINEAR)
     return result
 
-def undistort_with_map_result(img, map_result):
+
+def undistort_with_map_result(img, map_result, crop=False, interpolation=cv2.INTER_LINEAR):
     assert 'map1' in map_result, f' !! map1 not in map_result: {map_result}.'
     assert 'map2' in map_result, f' !! map2 not in map_result: {map_result}.'
     
@@ -1968,7 +2023,12 @@ def undistort_with_map_result(img, map_result):
 
     assert map1.shape[:2] == img.shape[:2] == map2.shape[:2], f' !! shape not match: map1 {map1.shape[:2]}, map2 {map2.shape[:2]}, img {img.shape}.'
 
-    undistorted = cv2.remap(img, map1, map2, cv2.INTER_LINEAR)
+    undistorted = cv2.remap(img, map1, map2, interpolation)
+
+    if crop:
+        roi_bbox = map_result['roi_bbox']
+        undistorted = undistorted[roi_bbox[1]:roi_bbox[3], roi_bbox[0]:roi_bbox[2]]
+
     return undistorted
 
 
@@ -1992,3 +2052,69 @@ def colorful_mask(ori_output, color_dict=None):
         puzzle[ori_output==key] = color_dict[key]
 
     return puzzle.astype('uint8')
+
+
+def read_mongodb_image(mongo_img_file_obj):
+    image_stream = io.BytesIO(mongo_img_file_obj.read())
+
+    pil_image = Image.open(image_stream)
+    pil_image = pil_image.convert('RGB')
+    image = np.array(pil_image)                             # 5000 * 4000 * 3 uint8 图像耗时 0.17 s
+
+    return image
+
+
+def make_voronoi_grad_mask(width, height, voronoi_polygon, gradient_width):
+    voronoi_img = np.zeros((height, width), dtype=np.float32)
+    cv2.polylines(voronoi_img, voronoi_polygon[None, :, :].astype('int32'), isClosed=True, color=255, thickness=vvd_round(gradient_width))
+
+    Ys, Xs = np.where(voronoi_img > 0)
+    voronoi_img[...] = 0
+
+    cv2.fillPoly(voronoi_img, voronoi_polygon[None, :, :].astype('int32'), color=1)
+    cv2.erode(voronoi_img, np.ones((7, 7), dtype=np.uint8), voronoi_img)
+
+    for x, y in zip(Xs, Ys):
+        row, col = int(y), int(x)
+        distance = cv2.pointPolygonTest(voronoi_polygon, (col, row), True)
+        value = np.clip(0.5 + distance / gradient_width, 0, 1)
+        voronoi_img[row, col] = value
+
+    return voronoi_img
+
+
+def make_voronoi_mask(width, height, voronoi_polygon):
+    voronoi_img = np.zeros((height, width), dtype=np.uint8)
+    cv2.fillPoly(voronoi_img, voronoi_polygon[None, :, :].astype('int32'), color=1)
+
+    return voronoi_img
+
+
+class UndistortProcessor:
+    def __init__(self, fx, fy, cx, cy, k1, k2, p1, p2, k3, W, H):
+
+        mtx = np.array([[fx, 0, cx], [0, fy, cy], [0, 0, 1]])
+        dist = np.array([k1, k2, p1, p2, k3])
+
+        self.undistort_param_dict = distortion_calibration_to_map(mtx, dist, W, H)
+        self.W = W
+        self.H = H
+
+    def __call__(self, img, interpolation=cv2.INTER_LINEAR):
+        return self.process_img(img, interpolation=interpolation)
+
+    def process_img(self, img, crop=False, interpolation=cv2.INTER_LINEAR):
+        W, H = img.shape[1], img.shape[0]
+        assert W == self.W and H == self.H, f"Image size {W}x{H} is not equal to {self.W}x{self.H}"
+
+        undistort_image = undistort_with_map_result(img, self.undistort_param_dict, crop, interpolation)
+
+        return undistort_image
+
+    # data_processor = UndistortProcessor(3706.15, 3706.15, 2639.71, 1937.01, -0.110588, 0.0124022, -0.0001086, -0.000254551, -0.0262282, 5280, 3956)
+    # for image_path in vv.tqdm(img_path_list):
+
+    #     img = vv.cv_rgb_imread(image_path)
+        
+    #     processed_img = data_processor(img)
+    #     vv.PIS([img, 'ori_img'], [processed_img, 'undistorted'])
